"""Guide menu unit tests."""

##############################################################################
# Pytest imports.
from pytest import fixture

##############################################################################
# Library imports.
from ngdb import NortonGuide
from ngdb.menu import Menu

##############################################################################
# Local imports.
from . import BIG_GUIDE


##############################################################################
@fixture
def guide() -> NortonGuide:
    """The guide to test with."""
    return NortonGuide(BIG_GUIDE)


##############################################################################
def test_has_menus(guide: NortonGuide) -> None:
    """The test guide has the correct number of menus."""
    assert guide.menu_count == 1
    assert guide.menu_count == len(guide.menus)
    assert all(isinstance(menu, Menu) for menu in guide.menus)


##############################################################################
def test_menu_title(guide: NortonGuide) -> None:
    """The menu title should load correctly."""
    assert guide.menus[0].title == "OSLIB"
    assert guide.menus[0].title == str(guide.menus[0])


##############################################################################
def test_menu_options(guide: NortonGuide) -> None:
    """The menu options should load correctly."""
    assert guide.menus[0].prompts == (
        "Functions",
        "FAQs",
        "Revision History",
        "Credits",
        "About",
    )
    assert [prompt.text for prompt in guide.menus[0]] == [
        "Functions",
        "FAQs",
        "Revision History",
        "Credits",
        "About",
    ]


##############################################################################
def test_menu_item(guide: NortonGuide) -> None:
    """It should be possible to treat a menu like a list."""
    assert guide.menus[0][0][0] == "Functions"
    assert guide.menus[0][0].text == "Functions"
    assert str(guide.menus[0][0]) == "Functions"
    assert guide.menus[0][0][1] == 525
    assert guide.menus[0][0].offset == 525
    assert int(guide.menus[0][0]) == 525
    assert guide.menus[0][0]


### test_guide_menu.py ends here



from ngdb import Long, NortonGuide, Short

if __name__ == "__main__":
    with NortonGuide("tests/guides/oslib.ng") as guide:
        guide.goto_first()
        while not guide.eof:
            entry = guide.load()
            if isinstance(entry, Short):
                print("Short")
            elif isinstance(entry, Long):
                print("Long")
            else:
                print("This can't be possible")
            guide.skip()



"""Defines helpful types and values for the library.."""

##############################################################################
# Python imports.
from enum import Enum


##############################################################################
class NGDBError(Exception):
    """Base exception of all exceptions in the library."""


##############################################################################
class UnknownEntryType(NGDBError):
    """Type of an exception when faced with an unknown entry type."""


##############################################################################
class NGEOF(NGDBError):
    """Type of an exception thrown when doing things at or past EOF."""


##############################################################################
class EntryType(Enum):
    """Types of entry in a guide."""

    SHORT = 0
    """The record ID for a short entry in a Norton Guide database."""

    LONG = 1
    """The record ID for a long entry in a Norton Guide database."""

    MENU = 2
    """The record ID for a menu in a Norton Guide database."""

    @classmethod
    def is_short(cls, test: int) -> bool:
        """Is the value the ID of a short entry?

        Args:
            test: The value to test.

        Returns:
            ``True`` if it is a short, ``False`` if not.
        """
        return cls(test) is cls.SHORT

    @classmethod
    def is_long(cls, test: int) -> bool:
        """Is the value the ID of a long entry?

        Args:
            test: The value to test.

        Returns:
            ``True`` if it is a long, ``False`` if not.
        """
        return cls(test) is cls.LONG

    @classmethod
    def is_menu(cls, test: int) -> bool:
        """Is the value the ID of a menu?

        Args:
            test: The value to test.

        Returns:
            ``True`` if it is a menu, ``False`` if not.
        """
        return cls(test) is cls.MENU


### types.py ends here



"""Norton Guide parser for the text inside a guide."""

##############################################################################
# Python compatibility hackage.
from __future__ import annotations

##############################################################################
# Python imports.
from abc import ABC, abstractmethod
from enum import Enum, auto
from typing import Final

##############################################################################
# Local imports.
from .dosify import make_dos_like


##############################################################################
class TextMode(Enum):
    """Types of text mode used when parsing a Norton Guide line."""

    NORMAL = auto()
    """Normal text."""
    BOLD = auto()
    """Bold text."""
    UNDERLINE = auto()
    """Underlined text."""
    REVERSE = auto()
    """Reverse text."""
    ATTR = auto()
    """Raw colour attribute."""


##############################################################################
CTRL_CHAR: Final[str] = "^"
"""The control character that marks an upcoming attribute."""


##############################################################################
class ParseState:
    """Raw text parsing state tracking class."""

    def __init__(self, line: str) -> None:
        """Constructor.

        Args:
            line: The line to work on.
        """
        self.raw = line
        """The current raw text that's left to handle."""
        self.ctrl = line.find(CTRL_CHAR)
        """The location of the next control marker."""
        self.mode = TextMode.NORMAL
        """The current mode."""
        self.last_attr = -1
        """The last attribute encountered."""

    @property
    def work_left(self) -> bool:
        """Is there any work left to do?"""
        return self.ctrl != -1 and self.ctrl < len(self.raw)

    @property
    def ctrl_id(self) -> str:
        """The current control ID."""
        try:
            return self.raw[self.ctrl + 1].lower()
        except IndexError:
            # If we've fallen in here, it's mostly because we've run into
            # some situation where there's a lone ^ at the end of the line.
            # This feels like a detail that the user-level code should not
            # be having to faff with. I feel that, as much as possible, the
            # parsing code should do its absolute best to return something
            # readable when faced with invalid markup.
            return ""


##############################################################################
class BaseParser:
    """The base text parsing class."""

    def __init__(self, line: str) -> None:
        """Constructor.

        Args:
            line: The raw string to parse.
        """

        # State tracker.
        state = ParseState(line)

        # While we've not run out of text to process...
        while state.work_left:
            # If there was text between the last markup and the next...
            if len(state.raw[: state.ctrl]) > 0:
                # ...handle it.
                self.text(state.raw[: state.ctrl])

            # Pull out the character following the control character and
            # handle it.
            if (ctrl := state.ctrl_id) == CTRL_CHAR:
                # We're looking at ^^, which is a ^.
                self.text(CTRL_CHAR)
                state.ctrl += 2
            elif hasattr(self, f"_ctrl_{ctrl}"):
                # Looks like we can handle whatever's there, so dispatch
                # it...
                getattr(self, f"_ctrl_{ctrl}")(state)
            else:
                # No idea what the next character is. We could either raise
                # an exception, eat the next character, or simply skip along
                # one. For now, let's just skip along one.
                state.ctrl += 1

            # Chop the bits we've done off the raw string.
            state.raw = state.raw[state.ctrl :]

            # Find the next control character.
            state.ctrl = state.raw.find(CTRL_CHAR)

        # Handle any remaining text.
        if len(state.raw) > 0:
            self.text(state.raw)

    def _ctrl_a(self, state: ParseState) -> None:
        """Handle ^A markup.

        Args:
            state: The data that tracks parse state.
        """

        # Get the actual attribute.
        attr = int(state.raw[state.ctrl + 2 : state.ctrl + 4], 16)

        # If there's already a colour attribute in effect and the
        # new colour is the same as the previous colour...
        if state.mode is TextMode.ATTR and attr == state.last_attr:
            # ...that means it's a return to "normal".
            self.normal()
            state.mode = TextMode.NORMAL
        else:
            # ...otherwise we start a colour attribute.
            self.colour(attr)
            state.last_attr = attr
            state.mode = TextMode.ATTR

        # Skip.
        state.ctrl += 4

    def _ctrl_b(self, state: ParseState) -> None:
        """Handle ^B markup.

        Args:
            state: The data that tracks parse state.
        """

        # If we're in bold mode...
        if state.mode is TextMode.BOLD:
            # ...go back to normal.
            self.unbold()
            state.mode = TextMode.NORMAL
        else:
            # ...otherwise go bold!
            self.bold()
            state.mode = TextMode.BOLD

        # Skip!
        state.ctrl += 2

    def _ctrl_c(self, state: ParseState) -> None:
        """Handle ^C markup.

        Args:
            state: The data that tracks parse state.
        """
        self.char(int(state.raw[state.ctrl + 2 : state.ctrl + 4], 16))
        state.ctrl += 4

    def _ctrl_n(self, state: ParseState) -> None:
        """Handle ^N markup.

        Args:
            state: The data that tracks parse state.
        """
        self.normal()
        state.mode = TextMode.NORMAL
        state.ctrl += 2

    def _ctrl_r(self, state: ParseState) -> None:
        """Handle ^R markup.

        Args:
            state: The data that tracks parse state.
        """

        # If we're in reverse mode...
        if state.mode is TextMode.REVERSE:
            # ...go back to normal.
            self.unreverse()
            state.mode = TextMode.NORMAL
        else:
            # ...otherwise go reverse.
            self.reverse()
            state.mode = TextMode.REVERSE

        # Skip!
        state.ctrl += 2

    def _ctrl_u(self, state: ParseState) -> None:
        """Handle ^U markup.

        Args:
            state: The data that tracks parse state.
        """

        # If we're in underline mode...
        if state.mode is TextMode.UNDERLINE:
            # ...go back to normal.
            self.ununderline()
            state.mode = TextMode.NORMAL
        else:
            # ...otherwise go underline.
            self.underline()
            state.mode = TextMode.UNDERLINE

        # Skip!
        state.ctrl += 2

    def text(self, text: str) -> None:
        """Handle the given text.

        Args:
            text: The text to handle.
        """
        del text  # pragma: no cover

    def colour(self, colour: int) -> None:
        """Handle the given colour value.

        Args:
            colour: The colour value to handle.
        """
        del colour  # pragma: no cover

    def normal(self) -> None:
        """Handle being asked to go to normal mode."""

    def bold(self) -> None:
        """Handle being asked to go to bold mode."""

    def unbold(self) -> None:
        """Handle being asked to go out of bold mode."""

    def reverse(self) -> None:
        """Handle being asked to go to reverse mode."""

    def unreverse(self) -> None:
        """Handle being asked to go out of reverse mode."""

    def underline(self) -> None:
        """Handle being asked to go in underline mode."""

    def ununderline(self) -> None:
        """Handle being asked to go out of underline mode."""

    def char(self, char: int) -> None:
        """Handle an individual character value.

        Args:
            char: The character value to handle.
        """
        del char  # pragma: no cover


##############################################################################
class PlainText(BaseParser):
    """Read a line of Norton Guide text as plain text."""

    def __init__(self, line: str) -> None:
        # We're going to accumulate the text into a hidden instance variable.
        self._text = ""
        # Having set the above up, go parse.
        super().__init__(line)

    def text(self, text: str) -> None:
        self._text += text

    def char(self, char: int) -> None:
        self.text(chr(char))

    def __str__(self) -> str:
        """Return the plain text of the line.

        Returns:
            The parsed line, as plan text.
        """
        return self._text


##############################################################################
class MarkupText(PlainText, ABC):
    """Read a line of Norton Guide text and mark up with start/end tags.

    This is an abstract base class for other parser classes that will
    implement start and end tags where necessary.
    """

    def __init__(self, line: str) -> None:
        # We're going to keep a stack of the markup.
        self._stack: list[str] = []
        # Having set the above up, go parse.
        super().__init__(line)

    @abstractmethod
    def open_markup(self, cls: str) -> str:
        """Open markup for the given class.

        Args:
            cls: The class of thing to open the markup for.

        Returns:
            The opening markup text.
        """
        return ""

    @abstractmethod
    def close_markup(self, cls: str) -> str:
        """Close markup for the given class.

        Args:
            cls: The class of thing to close the markup for.

        Returns:
            The closing markup text.
        """
        return ""

    def begin_markup(self, cls: str) -> None:
        """Start a section of markup.

        Args:
            cls: The class for the markup.

        Note:
            As a side-effect of calling on this, the ``close_markup`` for
            the same class will be placed on an internal stack, for use when
            ``end_markup`` is called.
        """
        self._text += self.open_markup(cls)
        self._stack.append(self.close_markup(cls))

    def end_markup(self) -> None:
        """End a section of markup."""
        self._text += self._stack.pop()

    def normal(self) -> None:
        """Handle being asked to go to normal mode.

        Note:
            Internally this also clears the whole stack of closing tags.
        """
        self._text += "".join(reversed(self._stack))
        self._stack = []

    def __str__(self) -> str:
        self.normal()
        return super().__str__()


##############################################################################
class RichText(MarkupText):
    """Read a line of Norton Guide text and mark up with Rich console markup.

    Note:
        This is implemented in a way that doesn't require that Rich is a
        dependency of this library. This is provided here as a test and a
        handy example, and one that uses [Rich's plain text BBCode-a-like
        markup](https://rich.readthedocs.io/en/stable/protocol.html).
    """

    def text(self, text: str) -> None:
        super().text(make_dos_like(text).replace("[", r"\["))

    def open_markup(self, cls: str) -> str:
        return f"[{cls}]"

    def close_markup(self, cls: str) -> str:
        del cls
        return "[/]"

    COLOUR_MAP: Final[dict[int, int]] = {
        1: 4,
        3: 6,
        4: 1,
        6: 3,
        9: 21,
        11: 14,
        12: 196,
        14: 11,
    }
    """DOS to Rich colour mapping. This is just the exceptions."""

    @classmethod
    def map_colour(cls, colour: int) -> int:
        """Map a DOS colour into a similar colour from Rich.

        Args:
            colour: The DOS colour to map from.

        Returns:
            The mapped colour.
        """
        return cls.COLOUR_MAP.get(colour, colour)

    def colour(self, colour: int) -> None:
        self.begin_markup(
            f"color({self.map_colour(colour & 0xF)}) on color({self.map_colour(colour >> 4 & 0xF)})"
        )

    def bold(self) -> None:
        self.begin_markup("bold")

    def unbold(self) -> None:
        self.end_markup()

    def reverse(self) -> None:
        self.begin_markup("reverse")

    def unreverse(self) -> None:
        self.end_markup()

    def underline(self) -> None:
        self.begin_markup("underline")

    def ununderline(self) -> None:
        self.end_markup()


### parser.py ends here



"""Unit tests relating to basic guide navigation."""

##############################################################################
# Pytest imports.
from pytest import fixture, raises

##############################################################################
# Library imports.
from ngdb import NGEOF, Long, NortonGuide, Short

##############################################################################
# Local imports.
from . import BIG_GUIDE, GOOD_GUIDE


##############################################################################
@fixture
def big_guide() -> NortonGuide:
    """A guide to test with."""
    return NortonGuide(BIG_GUIDE)


##############################################################################
@fixture
def good_guide() -> NortonGuide:
    """A guide to test with."""
    return NortonGuide(GOOD_GUIDE)


##############################################################################
def test_go_first(big_guide: NortonGuide) -> None:
    """It should be possible to go to the first entry."""
    assert isinstance(big_guide.goto_first().load(), Short)


##############################################################################
def test_skip(big_guide: NortonGuide) -> None:
    """It should be possible to skip an entry without reading it."""
    assert isinstance(big_guide.goto_first().skip().load(), Long)


##############################################################################
def test_small_eof_skip(good_guide: NortonGuide) -> None:
    """A guide with one entry should EOF when skipping."""
    assert good_guide.skip().eof is True


##############################################################################
def test_small_eof_load(good_guide: NortonGuide) -> None:
    """A guide with one entry should not EOF when loading."""
    good_guide.load()
    assert good_guide.eof is False


##############################################################################
def test_big_eof_skip(big_guide: NortonGuide) -> None:
    """A guide with multiple entries should not be EOF early on during skips."""
    big_guide.goto_first()
    for _ in range(5):
        big_guide.skip()
        assert big_guide.eof is False


##############################################################################
def test_big_eof_load(big_guide: NortonGuide) -> None:
    """A guide with multiple entries should not be EOF early on during loads."""
    big_guide.goto_first()
    for _ in range(5):
        big_guide.load()
        big_guide.skip()
        assert big_guide.eof is False


##############################################################################
def test_small_eof_guard_skip(good_guide: NortonGuide) -> None:
    """Attempting to skip past the end of single-entry guide should throw an error."""
    good_guide.goto_first()
    with raises(NGEOF):
        for _ in range(100):
            good_guide.skip()


##############################################################################
def test_big_eof_guard_skip(big_guide: NortonGuide) -> None:
    """Attempting to skip past the end of multiple-entry guide should throw an error."""
    big_guide.goto_first()
    with raises(NGEOF):
        for _ in range(100):
            big_guide.skip()


##############################################################################
def test_iter_small(good_guide: NortonGuide) -> None:
    """It should be possible to iterate through a guide with one entry."""
    assert len(list(good_guide)) == 1


##############################################################################
def test_iter_big(big_guide: NortonGuide) -> None:
    """It should be possible to iterate through a guide with more than one entry."""
    assert len(list(big_guide)) == 28


### test_navigation.py ends here



"""Provides a class that is the base for a prompt collection."""

##############################################################################
# Python compatibility hackage.
from __future__ import annotations

##############################################################################
# Python imports.
from typing import Final, Iterator

##############################################################################
# Local imports.
from .link import Link


##############################################################################
class PromptCollection:
    """Base class for classes that contain prompt/offset collections."""

    MAX_PROMPT_LENGTH: Final[int] = 128
    """The maximum length of a prompt in a guide."""

    def __init__(self) -> None:
        """Constructor."""
        self._count = 0
        self._prompts: tuple[str, ...] = ()
        self._offsets: tuple[int, ...] = ()

    @property
    def prompts(self) -> tuple[str, ...]:
        """The prompts in the collection."""
        return self._prompts

    @property
    def offsets(self) -> tuple[int, ...]:
        """The offsets into the guide for each prompt."""
        return self._offsets

    def __len__(self) -> int:
        """Get the number of prompts in the collection.

        Returns:
            The count of prompts.
        """
        return self._count

    def __getitem__(self, prompt: int) -> Link:
        """Get a prompt/offset pair.

        Args:
            prompt: The prompt to get the details for.

        Returns:
            The requested prompt and its offset.
        """
        return Link(self.prompts[prompt], self.offsets[prompt])

    def __iter__(self) -> Iterator[Link]:
        """Get an iterator of prompt and offset pairs.

        Yields:
            A link containing the prompt/offset, from the collection.
        """
        return (Link(*prompt) for prompt in zip(self.prompts, self.offsets))

    def __bool__(self) -> bool:
        """Test if there are any prompts in the collection.

        Returns:
            [`True`][True] if there are prompts, [`False`][False] if not.
        """
        return bool(len(self))


### prompts.py ends here



"""Norton Guide entry tests."""

##############################################################################
# Pytest imports.
from pytest import fixture, raises

##############################################################################
# Library imports.
from ngdb import Entry, Long, NortonGuide, Short, UnknownEntryType
from ngdb.types import EntryType

##############################################################################
# Local imports.
from . import BIG_GUIDE


##############################################################################
@fixture
def guide() -> NortonGuide:
    """The guide to test with."""
    return NortonGuide(BIG_GUIDE)


##############################################################################
@fixture
def short(guide: NortonGuide) -> Short:
    """A short guide entry to test against."""
    assert isinstance(entry := guide.goto_first().load(), Short)
    return entry


##############################################################################
@fixture
def long(guide: NortonGuide) -> Long:
    """A long guide entry to test against."""
    assert isinstance(entry := guide.goto_first().skip().load(), Long)
    return entry


##############################################################################
def test_unknown(guide: NortonGuide) -> None:
    """Attempting to load an unknown entry type should result in an exception."""
    with raises(UnknownEntryType):
        guide.goto(0).load()


##############################################################################
def test_short_load_correct_type(short: Entry) -> None:
    """A short entry should load as the correct type."""
    assert isinstance(short, Short)


##############################################################################
def test_short_correct_id(short: Entry) -> None:
    """A short entry should have a short entry ID."""
    assert short.type_id == 0
    assert EntryType.is_short(short.type_id) is True
    assert EntryType.is_long(short.type_id) is False


##############################################################################
def test_short_entry_size(short: Entry) -> None:
    """The test short entry should have the correct size."""
    assert short.size == 838


##############################################################################
def test_short_parent(short: Entry) -> None:
    """The test short entry should not have a parent."""
    assert bool(short.parent) is False


##############################################################################
def test_short_parent_line(short: Entry) -> None:
    """It test should entry should not have a parent line."""
    assert short.parent.has_line is False


##############################################################################
def test_short_parent_menu(short: Entry) -> None:
    """The test short entry should have a parent menu."""
    assert short.parent.has_menu is True


##############################################################################
def test_short_parent_prompt(short: Entry) -> None:
    """The test short entry should have a parent menu prompt."""
    assert short.parent.has_prompt is True


##############################################################################
def test_short_previous(short: Entry) -> None:
    """The test short should not have a previous entry."""
    assert short.has_previous is False


##############################################################################
def test_short_next(short: Entry) -> None:
    """The test short should not have a next entry."""
    assert short.has_next is False


##############################################################################
def test_short_str_entry(short: Entry) -> None:
    """The str() of the short entry should be the main text."""
    assert len(str(short).split("\n")) == len(short)


##############################################################################
def test_short_lines_and_offsets(short: Short) -> None:
    """There should be equal numbers of lines and offsets for the short."""
    assert len(short.lines) == len(short.offsets)


##############################################################################
def test_short_list_like(short: Short) -> None:
    """It should be possible to treat a short entry like a list."""
    assert short[0] == (
        " OL_95AppTitle()          Set/get the Windows 95 application title.",
        1389,
    )
    assert (
        short[0].text
        == " OL_95AppTitle()          Set/get the Windows 95 application title."
    )
    assert short[0].offset == 1389
    assert short[0].has_offset is True


##############################################################################
def test_iter_iter(short: Short) -> None:
    """It should be possible to treat a short entry like an iterator."""
    assert next(iter(short)) == (
        " OL_95AppTitle()          Set/get the Windows 95 application title.",
        1389,
    )


##############################################################################
def test_long_load_correct_type(long: Long) -> None:
    """A long entry should load as the correct type."""
    assert isinstance(long, Long)


##############################################################################
def test_long_correct_id(long: Long) -> None:
    """A long entry should have a long entry ID."""
    assert long.type_id == 1
    assert EntryType.is_short(long.type_id) is False
    assert EntryType.is_long(long.type_id) is True


##############################################################################
def test_long_entry_size(long: Long) -> None:
    """The long entry should have the correct size."""
    assert long.size == 940


##############################################################################
def test_long_parent(long: Long) -> None:
    """The long entry should have a parent."""
    assert bool(long.parent) is True


##############################################################################
def test_long_parent_line(long: Long) -> None:
    """The long entry should have a parent line."""
    assert long.parent.has_line is True
    assert long.parent.line == 0


##############################################################################
def test_long_parent_menu(long: Long) -> None:
    """The test long entry should have a parent menu."""
    assert long.parent.has_menu is True


##############################################################################
def test_long_parent_prompt(long: Long) -> None:
    """The test long entry should have a parent menu prompt."""
    assert long.parent.has_prompt is True


##############################################################################
def test_long_previous(long: Long) -> None:
    """The test long entry should not have a previous entry."""
    assert long.has_previous is False


##############################################################################
def test_long_next(long: Long) -> None:
    """The test long entry should have a next entry."""
    assert long.has_next is True


##############################################################################
def test_long_str_entry(long: Long) -> None:
    """The str() of the test long entry should be the main text."""
    assert len((str_entry := str(long)).split("\n")) == len(long)


##############################################################################
def test_long_list_like(long: Long) -> None:
    """It should be possible to treat a long entry like a list."""
    assert long[0] == " ^bOL_95AppTitle()"


##############################################################################
def test_long_iter_like(long: Long) -> None:
    """It should be possible to treat a long entry like an iterator."""
    assert next(iter(long)) == " ^bOL_95AppTitle()"


##############################################################################
def test_long_see_also(long: Long) -> None:
    """The test long entry should have a see-also menu."""
    assert bool(long.see_also)


##############################################################################
def test_long_see_also_count(long: Long) -> None:
    """The test long entry should have the correct number of see-also items."""
    assert len(long.see_also) == 1


##############################################################################
def test_long_see_also_text(long: Long) -> None:
    """The test long entry should have the correct see-also prompt."""
    assert long.see_also.prompts[0] == "OL_95VMTitle()"


##############################################################################
def test_long_see_also_offset(long: Long) -> None:
    """The test long entry should have the correct see-also offset."""
    assert long.see_also.offsets[0] == 2355


##############################################################################
def test_long_see_also_list_like(long: Long) -> None:
    """It should be possible to treat the see-also object like a list."""
    assert long.see_also[0] == ("OL_95VMTitle()", 2355)


##############################################################################
def test_long_see_also_iter_like(long: Long) -> None:
    """It should be possible to treat the see-also object like an iterator."""
    assert next(iter(long.see_also)) == ("OL_95VMTitle()", 2355)


### test_guide_entry.py ends here



"""Class for handling the low-level reading of a Norton Guide database."""

##############################################################################
# Python imports.
import io
from pathlib import Path
from typing import Final

##############################################################################
# Typing backward compatibility.
from typing_extensions import Self


##############################################################################
class GuideReader:
    """Low-level guide reading class.

    Note:
        For now, no optimisation has taken place, in many cases the way this
        class reads data from a guide will be a method that's on the slower
        side. This is on purpose; it's about readable code that represents
        the underlying data structure rather than the fastest method of
        getting data into memory.

        Once the rest of the library is done and working well this extra bit
        if docstring will likely be removed because work to improve the
        speed of this class will finally take place.
    """

    RLE_MARKER: Final[str] = chr(0xFF)
    """The value that marks run-length-encoded spaces."""

    @classmethod
    def unrle(cls, rle_text: str) -> str:
        """Un-run-length-encode the given string.

        Args:
            rle_text: The text that needs expanding.

        Returns:
            The given text with all RLE components expanded.

        Note:
            Norton Guide database files use a very simple form of
            run-length-encoding for spaces. Simply put, if you find a byte
            in a string that is `0xFF`, then the next byte is the number of
            spaces to insert into the string at this point. I've also found
            that `0xFF` followed by `0xFF` seems to mean you should insert a
            literal `0xFF` (I think), although I've found that using a
            literal space makes more sense.
        """

        expanded = ""
        start = 0
        split = rle_text.find(cls.RLE_MARKER)

        while split > -1:
            expanded += rle_text[start:split] + " " * (
                1 if rle_text[split + 1] == cls.RLE_MARKER else ord(rle_text[split + 1])
            )
            start = split + 2
            split = rle_text.find(cls.RLE_MARKER, start)

        return expanded + rle_text[start:]

    def __init__(self, guide: Path):
        """Constructor.

        Args:
            guide: The guide to open.
        """
        self._h = guide.open("rb")
        """The handle for the open guide."""

    def close(self) -> None:
        """Close the guide."""
        self._h.close()

    @property
    def pos(self) -> int:
        """The current position within the file."""
        return self._h.tell()

    def goto(self, pos: int) -> Self:
        """Go to a specific byte position within the guide.

        Args:
            pos: The position to go to.

        Returns:
            self
        """
        self._h.seek(pos)
        return self

    @property
    def closed(self) -> bool:
        """Is the file closed?"""
        return self._h.closed

    def skip(self, count: int = 1) -> Self:
        """Skip a number of bytes in the guide.

        Args:
            count: The optional number of bytes to skip.

        Returns:
            Self.
        """
        self._h.seek(count, io.SEEK_CUR)
        return self

    def skip_entry(self) -> Self:
        """Skip a whole entry in the guide.

        Returns:
            Self.
        """
        return self.skip(2).skip(self.read_word() + 22)

    @staticmethod
    def _decrypt(value: int) -> int:
        """Decrypt a given numeric value.

        Args:
            value: The value to decrypt.

        Returns:
            The decrypted value.
        """
        return value ^ 0x1A

    def read_byte(self, decrypt: bool = True) -> int:
        """Read a byte from the guide.

        Args:
            decrypt: Should the value be decrypted?

        Returns:
            The byte value read.
        """
        buff = self._h.read(1)[0]
        return self._decrypt(buff) if decrypt else buff

    def read_word(self, decrypt: bool = True) -> int:
        """Read a two-byte word from the guide.

        Args:
            decrypt: Should the value be decrypted?

        Returns:
            The word value read.
        """
        return self.read_byte(decrypt) + (self.read_byte(decrypt) << 8)

    def peek_word(self, decrypt: bool = True) -> int:
        """Read a two-byte word but don't move the file location.

        Args:
            decrypt: Should the value be decrypted?

        Returns:
            The word value read.
        """
        try:
            return self.read_word(decrypt)
        finally:
            self.skip(-2)

    def read_long(self, decrypt: bool = True) -> int:
        """Read a four-byte long word from the guide.

        Args:
            decrypt: Should the value be decrypted?

        Returns:
            The long integer value read.
        """
        return self.read_word(decrypt) + (self.read_word(decrypt) << 16)

    def read_offset(self) -> int:
        """Read an offset value from the guide.

        Returns:
            The offset value read.

        Note:
            This function ensures that an offset value that means 'there is
            no offset' returns as `-1`.
        """
        return -1 if (offset := self.read_long(True)) == 0xFFFFFFFF else offset

    @staticmethod
    def _nul_trim(string: str) -> str:
        """Trim a string from the first nul.

        Args:
            string: The string to trim.

        Returns:
            Everything up to but not including the first nul.
        """
        return string[0:nul] if (nul := string.find("\000")) != -1 else string

    def read_str(self, length: int, decrypt: bool = True) -> str:
        """Read a fixed-length string from the guide.

        Args:
            length: The length of the string to read.
            decrypt: Should the string be decrypted?

        Returns:
            The string value read.
        """
        return self._nul_trim(
            "".join(
                chr(self._decrypt(n) if decrypt else n)
                for n in tuple(self._h.read(length))
            )
        )

    def read_strz(self, length: int, decrypt: bool = True) -> str:
        """Read a nul-terminated string from the guide.

        This is similar to [`read_str`][ngdb.reader.GuideReader.read_str],
        but it will read only as far as the first `nul` it encounters,
        within the bounds of `length`, and the file read location will be
        correctly settled to take this into account.

        Args:
            length: The maximum length of the string to read.
            decrypt: Should the string be decrypted?

        Returns:
            The string value read.
        """
        # Remember where we are before we read in the string.
        pos = self.pos
        # Now read in the string.
        buff = self.read_str(length, decrypt)
        # Now skip to the location where the nul was found.
        self.goto(pos + len(buff) + 1)
        # Return the string.
        return buff


### reader.py ends here



"""Provides a class that links some text to an offset in a guide."""

##############################################################################
# Python imports.
from typing import NamedTuple


##############################################################################
class Link(NamedTuple):
    """A link within a Norton Guide, comprising of some text and an offset."""

    text: str
    """The text of the link."""

    offset: int
    """The offset of the link."""

    def __str__(self) -> str:
        """Returns the text of the link.

        Returns:
            The text for the link.
        """
        return self.text

    def __int__(self) -> int:
        """Returns the offset of the link.

        Returns:
            The offset for the link.
        """
        return self.offset

    @property
    def has_offset(self) -> bool:
        """Does this link have an associated offset into the guide?"""
        return self.offset > 0

    def __bool__(self) -> bool:
        """Does the link actually link anywhere?

        Returns:
            ``True`` if the link links somewhere, ``False`` if not.
        """
        return self.has_offset


### link.py ends here



"""Norton Guide entry content markup unit tests."""

##############################################################################
# Python compatibility hackage.
from __future__ import annotations

##############################################################################
# Python imports.
from typing import Iterator

##############################################################################
# Backward compatibility.
from typing_extensions import TypeAlias

##############################################################################
# Library imports.
from ngdb import BaseParser, PlainText
from ngdb.parser import RichText


##############################################################################
class TestPlainText:
    """Perform parser tests against the plain text parser."""

    def test_empty_string(self) -> None:
        """It can handle an empty line."""
        assert str(PlainText("")) == ""

    def test_no_markup(self) -> None:
        """It can handle a line with no markup."""
        assert str(PlainText("Test")) == "Test"

    def test_strips_non_text_markup(self) -> None:
        """It should strip all non-text markup without issue."""
        assert (
            str(PlainText("^A10This ^Bis ^Na ^Rtest, ^Ureally it is."))
            == "This is a test, really it is."
        )

    def test_strips_only_non_text_markup(self) -> None:
        """It should turn non-text markup only into an empty string."""
        assert str(PlainText("^A10^B^N^R^U")) == ""

    def test_ctrl_ctrl(self) -> None:
        """^^ should become ^."""
        assert str(PlainText("^^" * 10)) == ("^" * 10)

    def test_char_code(self) -> None:
        """It should be able to handle character codes."""
        assert str(PlainText("^C20^C21")) == " !"

    def test_truncated_markup(self) -> None:
        """It should handle truncated markup."""
        assert str(PlainText("^")) == ""


##############################################################################
TEvent: TypeAlias = str | tuple[str, str | int]
"""Type of a single event that the TestParser will catch."""

##############################################################################
TEvents: TypeAlias = list[TEvent]
"""Type of the collection of events in the TestParser."""


##############################################################################
# Unit-test-oriented Norton Guide line parser.
class MockedParser(BaseParser):
    """Parser class for working with unit tests."""

    def __init__(self, line: str) -> None:
        # First off, we're going to collect all the different events that
        # happen, so start a list for doing that.
        self._events: TEvents = []

        # Then call the super.
        super().__init__(line)

    def text(self, text: str) -> None:
        self._events.append(("T", text))

    def colour(self, colour: int) -> None:
        self._events.append(("A", colour))

    def normal(self) -> None:
        self._events.append("N")

    def bold(self) -> None:
        self._events.append("B")

    def unbold(self) -> None:
        self._events.append("b")

    def reverse(self) -> None:
        self._events.append("R")

    def unreverse(self) -> None:
        self._events.append("r")

    def underline(self) -> None:
        self._events.append("U")

    def ununderline(self) -> None:
        self._events.append("u")

    def char(self, char: int) -> None:
        self._events.append(("C", char))

    def __iter__(self) -> Iterator[TEvent]:
        """The collection of events caught by the parser."""
        return iter(self._events)


##############################################################################
class TestParseEvents:
    """Test the various events that happen within a parser."""

    def test_empty_line(self) -> None:
        """There should be no events in an empty line."""
        assert list(MockedParser("")) == []

    def test_no_markup(self) -> None:
        """There should be a single text event for a non-markup line."""
        assert list(MockedParser("Hello, World!")) == [("T", "Hello, World!")]

    def test_colour(self) -> None:
        """There should be a colour event when there's a ^A."""
        assert list(MockedParser("Hello, ^A20World!")) == [
            ("T", "Hello, "),
            ("A", 0x20),
            ("T", "World!"),
        ]

    def test_multi_colour(self) -> None:
        """Multiple there should be multiple colour events with multiple ^A."""
        assert list(MockedParser("Hello, ^A20World^A64!")) == [
            ("T", "Hello, "),
            ("A", 0x20),
            ("T", "World"),
            ("A", 0x64),
            ("T", "!"),
        ]

    def test_same_colour(self) -> None:
        """Two consecutive ^A of the same colour should cause a ^N."""
        assert list(MockedParser("Hello, ^A20World^A20!")) == [
            ("T", "Hello, "),
            ("A", 0x20),
            ("T", "World"),
            "N",
            ("T", "!"),
        ]

    def test_bold(self) -> None:
        """It should be possible to turn bold on and off with ^B."""
        assert list(MockedParser("Hello, ^BWorld^B!")) == [
            ("T", "Hello, "),
            "B",
            ("T", "World"),
            "b",
            ("T", "!"),
        ]

    def test_char(self) -> None:
        """It should be possible to generate characters with ^C."""
        assert list(MockedParser("".join(f"^C{n:02x}" for n in range(256)))) == [
            ("C", n) for n in range(256)
        ]

    def test_normal(self) -> None:
        """A ^N markup should result in a back-to-normal event."""
        assert list(MockedParser("Hello, ^NWorld!")) == [
            ("T", "Hello, "),
            "N",
            ("T", "World!"),
        ]

    def test_reverse(self) -> None:
        """It should be possible to turn reverse on and off with ^R."""
        assert list(MockedParser("Hello, ^RWorld^R!")) == [
            ("T", "Hello, "),
            "R",
            ("T", "World"),
            "r",
            ("T", "!"),
        ]

    def test_underline(self) -> None:
        """It should be possible to turn underline on and off with ^U."""
        assert list(MockedParser("Hello, ^UWorld^U!")) == [
            ("T", "Hello, "),
            "U",
            ("T", "World"),
            "u",
            ("T", "!"),
        ]


##############################################################################
class TestRichParser:
    """Test the Rich-friendly parser."""

    def test_empty_line(self) -> None:
        """An empty line should parse fine."""
        assert str(RichText("")) == ""

    def test_no_markup(self) -> None:
        """A line with no markup should parse fine."""
        assert str(RichText("Hello, World!")) == "Hello, World!"

    def test_escape_markup(self) -> None:
        """Text that looks like Rich markup should get escaped."""
        assert str(RichText("Hello, [W]orld!")) == "Hello, \\[W]orld!"

    def test_colour(self) -> None:
        """A colour attribute should come out as the expected markup."""
        assert str(RichText("^A02Hello")) == "[color(2) on color(0)]Hello[/]"

    def test_bold(self) -> None:
        """A bold attribute should turn into the Rich version."""
        assert str(RichText("^BHello^b")) == "[bold]Hello[/]"

    def test_reverse(self) -> None:
        """A reverse attribute should turn into the Rich version."""
        assert str(RichText("^RHello^r")) == "[reverse]Hello[/]"

    def test_underline(self) -> None:
        """An underline attribute should turn into the Rich version."""
        assert str(RichText("^UHello^u")) == "[underline]Hello[/]"


### test_parser.py ends here



"""Norton guide entry loading and holding code."""

##############################################################################
# Python compatibility hackage.
from __future__ import annotations

##############################################################################
# Python imports.
from typing import Final, Iterator

##############################################################################
# Local imports.
from .link import Link
from .parser import RichText
from .reader import GuideReader
from .seealso import SeeAlso


##############################################################################
class EntryParent:
    """Class to load and hold the parent information for an entry."""

    def __init__(self, guide: GuideReader) -> None:
        """Constructor.

        Args:
            guide: The reader object for the guide.
        """
        self._line = guide.read_word()
        self._offset = guide.read_offset()
        self._menu = guide.read_word()
        self._prompt = guide.read_word()

    @property
    def offset(self) -> int:
        """The offset of the parent entry, if there is one."""
        return self._offset

    def __bool__(self) -> bool:
        """Is there a parent entry?

        Returns:
            [`True`][True] if there is, [`False`][False] if not.
        """
        return self.offset > 0

    @staticmethod
    def _non_test(value: int) -> int:
        """Ensure a -1 is a -1.

        Args:
            value: The value to clean up.

        Returns:
            The value with an intended -1 guaranteed.
        """
        return -1 if value == 0xFFFF else value

    @property
    def line(self) -> int:
        """The line in the parent entry that point to this entry.

        If there is no parent line this will be ``-1``. But also see
        [`has_line`][ngdb.entry.EntryParent.has_line] for a test for a
        parent entry line.
        """
        return self._non_test(self._line)

    @property
    def has_line(self) -> int:
        """Does this entry have a parent entry line that points to it?"""
        return self.line != -1

    @property
    def menu(self) -> int:
        """The menu that relates to this entry.

        If there is no menu, this will be `-1`. But also see
        [`has_menu`][ngdb.entry.EntryParent.has_menu] to test if there is a
        related menu.
        """
        return self._non_test(self._menu)

    @property
    def has_menu(self) -> bool:
        """Is there a menu related to this entry?"""
        return self.menu != -1

    @property
    def prompt(self) -> int:
        """The menu prompt related to this entry.

        If there is no menu prompt, this will be `-1`. But also see
        [`has_prompt`][ngdb.entry.EntryParent.has_prompt] to test if there
        is a related menu prompt.
        """
        return self._non_test(self._prompt)

    @property
    def has_prompt(self) -> bool:
        """Is there a menu prompt related to this entry?"""
        return self.has_menu and self.prompt != -1


##############################################################################
MAX_LINE_LENGTH: Final[int] = 1024
"""Maximum size of a line we'll look for in a guide."""


##############################################################################
class Entry:
    """Norton Guide database entry class."""

    def __init__(self, guide: GuideReader) -> None:
        """Constructor.

        Args:
            guide: The reader object for the guide.
        """

        # Load up the main details for the entry.
        self._offset = guide.pos
        self._type = guide.read_word()
        self._size = guide.read_word()
        self._line_count = guide.read_word()
        self._has_see_also = guide.read_word()
        self._parent = EntryParent(guide)
        self._previous = guide.read_offset()
        self._next = guide.read_offset()

        # Set up for loading in the lines.
        self._lines: tuple[str, ...] = ()

    def _load_lines(self, guide: GuideReader) -> None:
        """Load in all of the lines of text, from this point.

        Args:
            guide: The reader object for the guide.
        """
        self._lines = tuple(
            guide.unrle(guide.read_strz(MAX_LINE_LENGTH)) for _ in range(len(self))
        )

    @property
    def offset(self) -> int:
        """The file offset of this entry."""
        return self._offset

    @property
    def type_id(self) -> int:
        """The numeric ID of the type of entry."""
        return self._type

    @property
    def size(self) -> int:
        """The size of the entry in bytes."""
        return self._size

    def __len__(self) -> int:
        """The number of lines in the entry.

        Returns:
            The number of lines.
        """
        return self._line_count

    @property
    def has_see_also(self) -> bool:
        """Does this entry have any see-also items?"""
        return self._has_see_also > 0

    @property
    def parent(self) -> EntryParent:
        """Returns the parent entry information."""
        return self._parent

    @property
    def previous(self) -> int:
        """The location of the previous entry."""
        return self._previous

    @property
    def has_previous(self) -> bool:
        """Is there a previous entry?"""
        return self.previous > 0

    @property
    def next(self) -> int:
        """The location of the next entry."""
        return self._next

    @property
    def has_next(self) -> bool:
        """Is there a next entry?"""
        return self.next > 0

    @property
    def lines(self) -> tuple[str, ...]:
        """The lines of text in the entry."""
        return self._lines

    def __str__(self) -> str:
        """Return the text of the entry as a single string.

        Returns:
            The entry's text.
        """
        return "\n".join(self.lines)

    def __repr__(self) -> str:
        """Returns a string representation of the object.

        Returns:
            The name of the type of entry.
        """
        return f"<{self.__class__.__name__}: {self.offset}>"

    def __rich__(self) -> str:
        """Support being printed in a [Rich-enhanced REPL](https://rich.readthedocs.io/en/stable/protocol.html).

        Returns:
            Rich-friendly test.
        """
        return "\n".join(str(RichText(line)) for line in self.lines)


##############################################################################
class Short(Entry):
    """Short Norton Guide database entry."""

    def __init__(self, guide: GuideReader) -> None:
        """Constructor.

        Args:
            guide: The reader object for the guide.
        """
        super().__init__(guide)
        self._offsets = tuple(self._load_offsets(guide))
        """The offsets for each line in the short entry."""
        self._load_lines(guide)

    def _load_offsets(self, guide: GuideReader) -> Iterator[int]:
        """Load up the offsets for each of the lines in the entry.

        Args:
            guide: The reader object for the guide.

        Yields:
            The offset for each line in the entry.
        """
        for _ in range(len(self)):
            # Skip a word -- I don't know what this is.
            guide.skip(2)
            # Read the offset of the line.
            yield guide.read_offset()

    @property
    def offsets(self) -> tuple[int, ...]:
        """The offsets for each of the lines in the entry."""
        return self._offsets

    def __getitem__(self, line: int) -> Link:
        """Get a line and its offset.

        Args:
            line: The line to get the link information for.

        Returns:
            The link associated with that line.
        """
        return Link(self.lines[line], self.offsets[line])

    def __iter__(self) -> Iterator[Link]:
        """The lines in the entry along with the offsets into the guide.

        Returns:
            An iterator of link data.
        """
        return (Link(*line) for line in zip(self.lines, self.offsets))


##############################################################################
class Long(Entry):
    """Long Norton Guide database entry."""

    def __init__(self, guide: GuideReader) -> None:
        """Constructor.

        Args:
            guide: The reader object for the guide.
        """
        super().__init__(guide)
        self._load_lines(guide)
        self._see_also = SeeAlso(guide, self.has_see_also)
        """The see-also data for the long entry."""

    @property
    def see_also(self) -> SeeAlso:
        """The see-also information for this entry."""
        return self._see_also

    def __getitem__(self, line: int) -> str:
        """Get a line from the entry.

        Args:
            line: The lint to get.

        Returns:
            The line.
        """
        return self.lines[line]

    def __iter__(self) -> Iterator[str]:
        """The lines in the entry.

        Returns:
            An iterator of strings that are the lines in the entry.
        """
        return iter(self.lines)


### entry.py ends here



"""Defines the class for opening and managing a Norton Guide database."""

##############################################################################
# Python compatibility hackage.
from __future__ import annotations

##############################################################################
# Python imports.
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Final, Iterator, TypeVar

##############################################################################
# Typing backward compatibility.
from typing_extensions import Self

##############################################################################
# Local imports.
from .entry import Long, Short
from .menu import Menu
from .reader import GuideReader
from .types import NGEOF, EntryType, UnknownEntryType

##############################################################################
EOFResult = TypeVar("EOFResult")
"""Return type of a method decorated with [`@not_eof`][ngdb.guide.not_eof]."""


##############################################################################
def not_eof(meth: Callable[..., EOFResult]) -> Callable[..., EOFResult]:
    """Decorator to ensure a guide isn't at EOF before executing a method.

    Args:
        meth: The method fo protect.

    Returns:
        The guard.

    This decorator is used as part of [`NortonGuide`][ngdb.NortonGuide], to
    decorate functions that should test for being
    [`eof`][ngdb.NortonGuide.eof] before the work of the method is done. If
    the guide is `eof` [`NGEOF`][ngdb.types.NGEOF] will be raised.
    """

    @wraps(meth)
    def _guard(self: NortonGuide, *args: Any, **kwargs: Any) -> EOFResult:
        """Guard the given method call."""
        if self.eof:
            raise NGEOF
        return meth(self, *args, **kwargs)

    return _guard


##############################################################################
class NortonGuide:
    """Norton Guide database wrapper class."""

    MAGIC: Final[dict[str, str]] = {"EH": "Expert Help", "NG": "Norton Guide"}
    """Lookup for valid database magic markers."""

    TITLE_LENGTH: Final[int] = 40
    """The length of a title in the header."""

    CREDIT_LENGTH: Final[int] = 66
    """The length of a line in the credits."""

    def __init__(self, guide: str | Path) -> None:
        """Constructor.

        Args:
            guide: The guide to open.
        """

        self._path = Path(guide)
        """The path to the guide."""

        # Attempt to open the guide. Note that we're going to hold it open
        # until we're asked to close it in the close method, so we also
        # nicely ask pylint to hush.
        self._guide = GuideReader(self._path)

        # Now, having opened it fine, read in the header.
        self._read_header()

        # Having read in the header, does it look like it's a Norton Guide
        # database we've been pointed at?
        if self.is_a:
            # Seems so. In that case sort the menus.
            self._menus = tuple(menu for menu in self._read_menus())

            # The number of menus should be correct at this point.
            assert len(self._menus) == self._menu_count

            # At this point we should be sat on top of the first entry, so
            # let's remember where that is.
            self._first_entry = self._guide.pos

    @property
    def path(self) -> Path:
        """The path to the guide."""
        return self._path

    def _read_header(self) -> None:
        """Read the header of the Norton Guide database."""

        # First two bytes are the magic.
        self._magic = self._guide.read_str(2, False)

        # Skip 4 bytes; to this day I'm not sure what they're for.
        self._guide.skip(4)

        # Read the count of menu options.
        self._menu_count = self._guide.read_word(False)

        # Read the title of the guide.
        self._title = self._guide.read_str(self.TITLE_LENGTH, False)

        # Read the credits for the guide.
        self._credits = tuple(
            self._guide.read_str(self.CREDIT_LENGTH, False) for _ in range(5)
        )

    def _read_menus(self) -> Iterator[Menu]:
        """Read the menus from the guide.

        Yields
            A menu from the guide.
        """
        while EntryType.is_menu(self._guide.peek_word()):
            yield Menu(self._guide)

    @property
    def is_open(self) -> bool:
        """Is the guide open?"""
        # Note that I first ensure that this instance actually does have a
        # `_guide` property as it's possible that an exception gets thrown
        # in the constructor and I have a __del__ method to ensure that the
        # handle is closed if it's open (see below). This means that it's
        # possible to fail to be fully constructed yet also asked to be
        # destructed.
        return hasattr(self, "_guide") and not self._guide.closed

    @property
    def is_a(self) -> bool:
        """Is the guide actually a Norton Guide database?"""
        return self.magic in self.MAGIC

    def close(self) -> None:
        """Close the guide, if it's open.

        Note:
            Closing the guide when it isn't open is a non-op. It's always
            safe to make this call.
        """
        if self.is_open:
            self._guide.close()

    def __enter__(self) -> Self:
        """Handle entry to context."""
        return self

    def __exit__(self, *_: Any) -> None:
        """Handle exit from context."""
        self.close()

    def __del__(self) -> None:
        """Ensure we close the handle to the guide if we're deleted."""
        self.close()

    @property
    def menu_count(self) -> int:
        """The count of menu options in the guide."""
        return self._menu_count

    @property
    def title(self) -> str:
        """The title of the guide."""
        return self._title

    @property
    def credits(self) -> tuple[str, ...]:
        """The credits for the guide."""
        return self._credits

    @property
    def magic(self) -> str:
        """The magic value for the guide.

        This tells us if the file is likely a Norton Guide database or not.
        It's always a two-character string and, normally, is ''NG''.
        However, if the guide was made for Expert Help, it could be ''EH''.
        """
        return self._magic

    @property
    def made_with(self) -> str:
        """The name of the tool that was used to make the guide."""
        return self.MAGIC.get(self.magic, "Unknown")

    @property
    def menus(self) -> tuple[Menu, ...]:
        """The menus for the guide."""
        return self._menus

    def goto(self, pos: int) -> Self:
        """Go to a specific location in the guide.

        Args:
            pos: The position to go to.

        Returns:
            Returns ``self``.
        """
        self._guide.goto(pos)
        return self

    def goto_first(self) -> Self:
        """Go to the first entry in the guide.

        Returns:
            Returns ``self``.
        """
        return self.goto(self._first_entry)

    @not_eof
    def skip(self) -> Self:
        """Skip the current entry.

        Returns:
            Returns ``self``.

        Raises:
            NGEOF: If we attempt to skip when at EOF.
        """
        self._guide.skip_entry()
        return self

    @property
    def eof(self) -> bool:
        """Are we at the end of the guide?"""
        return self._guide.pos >= self._path.stat().st_size

    @not_eof
    def load(self) -> Short | Long:
        """Load the entry at the current position.

        Returns:
            The entry found at the current position. Either a
              [`Short`][ngdb.Short] or a [`Long`][ngdb.Long] entry.

        Raises:
            NGEOF: If we attempt to load when at EOF.
            UnknownEntryType: If the type wasn't an entry.
        """
        pos = self._guide.pos
        try:
            match EntryType(self._guide.peek_word()):
                case EntryType.SHORT:
                    return Short(self._guide)
                case EntryType.LONG:
                    return Long(self._guide)
                case EntryType.MENU:
                    raise UnknownEntryType(
                        "Expecting a short or long entry but got a menu"
                    )
        except ValueError:
            raise UnknownEntryType(
                f"Unknown entry type: {self._guide.peek_word()}"
            ) from None
        finally:
            self.goto(pos)

    def __iter__(self) -> Iterator[Short | Long]:
        """Allow iterating through every entry in the guide.

        Yields:
            An entry from the guide. Either a [`Short`][ngdb.Short] or a
              [`Long`][ngdb.Long] entry.

        Raises:
            UnknownEntryType: If an unknown type of entry is encountered.
        """
        # Here I try my best to do this in a way that no other operations
        # that consume this iterator will affect it. So, starting with the
        # first entry...
        entry = self.goto_first().load()
        while True:
            try:
                # ...pass up the current entry...
                yield entry
                # ...and then, assuming the worst (that our caller may have
                # moved around the guide while consuming that entry), we
                # pointedly go back to it, skip it and load whatever's next.
                entry = self.goto(entry.offset).skip().load()
            except NGEOF:
                # EOF was thrown so let's finish the iterator.
                break

    def __repr__(self) -> str:
        """The string representation of the guide.

        Returns:
            The guide's full path/file name.
        """
        return f'<{self.__class__.__name__}: "{self}">'

    def __str__(self) -> str:
        """The string representation of the guide.

        Returns:
            The guide's full path/file name.
        """
        return str(self._path.resolve())


### guide.py ends here



"""ngdb - Norton Guide database reading library."""

##############################################################################
# Python imports.
from importlib.metadata import version

######################################################################
# Main library information.
__author__ = "Dave Pearson"
__copyright__ = "Copyright 2021-2025, Dave Pearson"
__credits__ = ["Dave Pearson"]
__maintainer__ = "Dave Pearson"
__email__ = "davep@davep.org"
__version__ = version("ngdb")
__licence__ = "GPLv3+"

##############################################################################
# Local imports.
from .dosify import make_dos_like
from .entry import Entry, Long, Short
from .guide import NortonGuide
from .link import Link
from .menu import Menu
from .parser import BaseParser, MarkupText, PlainText
from .seealso import SeeAlso
from .types import NGEOF, NGDBError, UnknownEntryType

##############################################################################
# Exports.
__all__ = (
    "BaseParser",
    "Entry",
    "Link",
    "Long",
    "make_dos_like",
    "MarkupText",
    "Menu",
    "NGDBError",
    "NGEOF",
    "NortonGuide",
    "PlainText",
    "SeeAlso",
    "Short",
    "UnknownEntryType",
)

### __init__.py ends here



from ngdb import NGEOF, NortonGuide

if __name__ == "__main__":
    with NortonGuide("tests/guides/oslib.ng") as guide:
        guide.goto_first()
        while True:
            print("Found an entry")
            try:
                guide.skip()
            except NGEOF:
                print("Hit the end of the guide!")
                break



from ngdb import NortonGuide

if __name__ == "__main__":
    with NortonGuide("tests/guides/oslib.ng") as guide:
        guide.goto_first()
        while not guide.eof:
            print("Found an entry")
            guide.skip()



"""Code for loading and holding guide menus."""

##############################################################################
# Local imports.
from .prompts import PromptCollection
from .reader import GuideReader


##############################################################################
class Menu(PromptCollection):
    """Class that loads and holds the details of a menu in the guide."""

    def __init__(self, guide: GuideReader) -> None:
        """Constructor.

        Args:
            guide: The reader object for the guide.
        """

        # Call the parent first.
        super().__init__()

        # Skip the type marker for the menu. Our caller should have tested
        # that we're a menu.
        _ = guide.read_word(False)

        # Skip the byte size of the menu section.
        _ = guide.read_word(False)

        # Next up, read the prompt count.
        self._count = guide.read_word() - 1

        # Now skip 20 bytes. I'm not sure what they are.
        guide.skip(20)

        # Next up is the collection of offsets for each menu prompt.
        self._offsets = tuple(guide.read_offset() for _ in range(len(self)))

        # Skip a number of values I don't know the purpose of, but I've
        # never needed. It seems to be two sets of long integer arrays.
        guide.skip((len(self) + 1) * 8)

        # We've now got to the title of the menu.
        self._title = guide.read_strz(self.MAX_PROMPT_LENGTH)

        # After the title comes the prompts.
        self._prompts = tuple(
            guide.read_strz(self.MAX_PROMPT_LENGTH) for _ in range(len(self))
        )

        # Finally, skip an unknown byte. This should then place us on the
        # next "record" in the database.
        guide.skip()

    @property
    def title(self) -> str:
        """The title of the menu."""
        return self._title

    def __str__(self) -> str:
        """Get the string representation of the menu.

        Returns:
            The title of the menu.
        """
        return self.title

    def __repr__(self) -> str:
        """Get the string representation of the menu.

        Returns:
            The title of the menu.
        """
        return f"<{self.__class__.__name__}: {self}>"


### menu.py ends here



"""See-also loading/holding code."""

##############################################################################
# Python imports.
from typing import Final

##############################################################################
# Local imports.
from .prompts import PromptCollection
from .reader import GuideReader


##############################################################################
class SeeAlso(PromptCollection):
    """Class to load and hold all the see alsos for a long entry."""

    MAX_SEE_ALSO: Final[int] = 20
    """Max number of see also items we'll handle.

    This is the limit published in the Expert Help Compiler manual and,
    while this limit isn't really needed in this code, it does help guard
    against corrupt guides.
    """

    def __init__(self, guide: GuideReader, load: bool) -> None:
        """Constructor.

        Args:
            guide: The reader object for the guide.
            load: Should we bother trying to load any?

        Note:
            The ``load`` parameter might look a bit daft, and it is, but a
            Norton Guide has a flag to say if there are any see-also entries
            *and* a count value later on, which won't be there and won't be
            0 if the flag is 0. So... we pass the flag in here and just have
            a see-also collection that has nothing in it.

            I could, of course, make the use of this class optional but that
            feels icky.
        """

        # Call the parent first.
        super().__init__()

        # Should we actually bother trying to load anything?
        if load:
            # Get the count of see-also entries.
            self._count = min(guide.read_word(), self.MAX_SEE_ALSO)

            # Get the offsets for each of the see-also entries.
            self._offsets = tuple(guide.read_offset() for _ in range(len(self)))

            # Get the prompts for each of the see-also items.
            self._prompts = tuple(
                guide.unrle(guide.read_strz(self.MAX_PROMPT_LENGTH))
                for _ in range(len(self))
            )


### seealso.py ends here
